use std::sync::Arc;

use vulkano::{
    command_buffer::allocator::StandardCommandBufferAllocator,
    descriptor_set::allocator::StandardDescriptorSetAllocator,
    device::{Device, Queue},
    format::Format,
    instance::Instance,
    memory::allocator::StandardMemoryAllocator,
};
use vulkano_util::renderer::VulkanoWindowRenderer;

use crate::core::{input::InputManager, render::vulkan_context::VulkanContext, timer::Timer};

pub struct SceneContext {
    pub instance: Arc<Instance>,
    pub device: Arc<Device>,
    pub graphics_queue: Arc<Queue>,
    pub compute_queue: Arc<Queue>,
    pub transfer_queue: Option<Arc<Queue>>,
    pub memory_allocator: Arc<StandardMemoryAllocator>,
    pub command_buffer_allocator: Arc<StandardCommandBufferAllocator>,
    pub descriptor_set_allocator: Arc<StandardDescriptorSetAllocator>,
    pub window_size: [f32; 2],
    pub aspect_ratio: f32,
    pub swapchain_format: Format,
    pub input_manager: Arc<InputManager>,
    pub timer: Arc<Timer>,
}

impl
    From<(
        &VulkanoWindowRenderer,
        &Arc<VulkanContext>,
        &Arc<InputManager>,
        &Arc<Timer>,
    )> for SceneContext
{
    fn from(
        (renderer, vulkan_context, input_manager, timer): (
            &VulkanoWindowRenderer,
            &Arc<VulkanContext>,
            &Arc<InputManager>,
            &Arc<Timer>,
        ),
    ) -> Self {
        let (command_buffer_allocator, descriptor_set_allocator) = {
            (
                vulkan_context.command_buffer_allocator().clone(),
                vulkan_context.descriptor_set_allocator().clone(),
            )
        };

        let (instance, device, graphics_queue, compute_queue, transfer_queue, memory_allocator) = {
            let vulkan_context = vulkan_context.vulkano_context();
            (
                vulkan_context.instance().clone(),
                vulkan_context.device().clone(),
                vulkan_context.graphics_queue().clone(),
                vulkan_context.compute_queue().clone(),
                vulkan_context.transfer_queue().cloned(),
                vulkan_context.memory_allocator().clone(),
            )
        };

        Self {
            instance,
            device,
            graphics_queue,
            compute_queue,
            transfer_queue,
            memory_allocator,
            command_buffer_allocator,
            descriptor_set_allocator,
            window_size: renderer.window_size(),
            aspect_ratio: renderer.aspect_ratio(),
            swapchain_format: renderer.swapchain_format(),
            input_manager: input_manager.clone(),
            timer: timer.clone(),
        }
    }
}
