use std::{error::Error, sync::Arc};

use crate::core::render::primitives::camera::Camera;
use crate::core::render::texture::Texture;
use crate::core::scene::Scene;
use crate::core::scene::SceneContext;
use egui_winit_vulkano::{Gui, egui};
use glam::Mat4;
use vulkano::{
    command_buffer::{
        AutoCommandBufferBuilder, CommandBufferUsage, PrimaryCommandBufferAbstract,
        RenderingAttachmentInfo, RenderingInfo,
    },
    image::view::ImageView,
    pipeline::graphics::viewport::Viewport,
    render_pass::{AttachmentLoadOp, AttachmentStoreOp},
    sync::GpuFuture,
};

use super::assets::square::Square;

pub struct MainSceneState {
    square: Square,
    camera: Camera,
    texture: Texture,
    speed: f32,
}

#[derive(Default)]
pub struct MainScene {
    state: Option<MainSceneState>,
}

impl Scene for MainScene {
    fn loaded(&self) -> bool {
        self.state.is_some()
    }

    fn load(&mut self, scene_context: &SceneContext) -> Result<(), Box<dyn Error>> {
        let square = Square::new(
            &scene_context.device,
            &scene_context.memory_allocator,
            scene_context.swapchain_format,
        )?;

        let camera = Camera::new(Mat4::perspective_rh_gl(
            std::f32::consts::FRAC_PI_2,
            scene_context.aspect_ratio,
            0.01,
            1000.0,
        ));

        let mut uploads = AutoCommandBufferBuilder::primary(
            scene_context.command_buffer_allocator.clone(),
            scene_context.graphics_queue.queue_family_index(),
            CommandBufferUsage::OneTimeSubmit,
        )?;

        let texture = Texture::from_file(
            &scene_context.device,
            &scene_context.memory_allocator,
            &mut uploads,
            "res/textures/wooden-crate.jpg",
        )?;

        let _ = uploads
            .build()?
            .execute(scene_context.graphics_queue.clone())?;

        self.state = Some(MainSceneState {
            square,
            camera,
            texture,
            speed: 50.0,
        });

        Ok(())
    }

    fn update(&mut self, scene_context: &SceneContext) -> Result<(), Box<dyn Error>> {
        let state = self.state.as_mut().ok_or("State not found")?;
        state.camera.update(
            &scene_context.input_manager,
            &scene_context.timer,
            state.speed,
            10.0,
        );

        Ok(())
    }

    fn render(
        &mut self,
        image_view: &Arc<ImageView>,
        acquire_future: Box<dyn GpuFuture>,
        scene_context: &SceneContext,
        gui: &mut Gui,
    ) -> Result<Box<dyn GpuFuture>, Box<dyn Error>> {
        let state = self.state.as_ref().ok_or("State not found")?;

        let mut builder = AutoCommandBufferBuilder::primary(
            scene_context.command_buffer_allocator.clone(),
            scene_context.graphics_queue.queue_family_index(),
            CommandBufferUsage::OneTimeSubmit,
        )?;

        {
            let viewport = Viewport {
                offset: [0.0, 0.0],
                extent: scene_context.window_size,
                depth_range: 0.0..=1.0,
            };

            builder
                .begin_rendering(RenderingInfo {
                    color_attachments: vec![Some(RenderingAttachmentInfo {
                        load_op: AttachmentLoadOp::Clear,
                        store_op: AttachmentStoreOp::Store,
                        clear_value: Some([0.0, 0.0, 0.0, 1.0].into()),
                        ..RenderingAttachmentInfo::image_view(image_view.clone())
                    })],
                    ..Default::default()
                })?
                .set_viewport(0, [viewport].into_iter().collect())?;
        }

        let camera_uniform = state
            .camera
            .create_buffer(&scene_context.memory_allocator)?;

        state
            .square
            .render(
                &mut builder,
                &scene_context.descriptor_set_allocator,
                &camera_uniform,
                &state.texture,
            )
            .unwrap();

        builder.end_rendering()?;

        let command_buffer = builder.build()?;

        let render_future =
            acquire_future.then_execute(scene_context.graphics_queue.clone(), command_buffer)?;

        gui.immediate_ui(|gui| {
            let ctx = gui.context();

            egui::Window::new("Informations")
                .vscroll(true)
                .show(&ctx, |ui| {
                    ui.label(format!("Resolution: {:?}", scene_context.window_size));

                    ui.label(format!("{:#?}", scene_context.input_manager));

                    ui.label(format!(
                        "Delta time: {:?}",
                        scene_context.timer.delta_time()
                    ));

                    ui.label(format!(
                        "Position: {:?}, Rotation: {:?}",
                        state.camera.get_position(),
                        state.camera.get_rotation()
                    ));
                });
        });

        let render_future = gui.draw_on_image(render_future, image_view.clone());

        Ok(render_future)
    }

    fn unload(&mut self) {}
}
